from functools import partial
from typing import Type

from sklearn.base import RegressorMixin

from asf.normalization.normalizations import AbstractNormalization, LogNormalization
from asf.predictors import SklearnWrapper
from asf.predictors.abstract_predictor import AbstractPredictor


class EPM:
    def __init__(
        self,
        predictor_class: Type[AbstractPredictor] | Type[RegressorMixin],
        normalization_class: Type[AbstractNormalization] = LogNormalization,
        transform_back: bool = True,
        predictor_config=None,
        predictor_kwargs=None,
    ):
        if isinstance(predictor_class, type) and issubclass(
            predictor_class, (RegressorMixin)
        ):
            self.model_class = partial(SklearnWrapper, predictor_class)
        else:
            self.model_class = predictor_class

        self.predictor_class = predictor_class
        self.normalization_class = normalization_class
        self.transform_back = transform_back
        self.predictor_config = predictor_config
        self.predictor_kwargs = predictor_kwargs or {}

    def fit(self, X, y, sample_weight=None):
        """
        Fit the EPM model to the data.

        Parameters:
        X: Features
        y: Target variable
        sample_weight: Sample weights (optional)
        """
        self.normalization = self.normalization_class()
        y = self.normalization.fit(y)

        if self.predictor_config is None:
            self.configuration_space = self.predictor_class()
        else:
            self.predictor = self.predictor_class.get_from_configuration(
                self.predictor_config, self.predictor_kwargs
            )

        self.predictor.fit(X, y, sample_weight=sample_weight)
        return self

    def predict(self, X):
        """
        Predict using the fitted EPM model.

        Parameters:
        X: Features"
        "
        """
        y_pred = self.predictor.predict(X)

        if self.transform_back:
            y_pred = self.normalization.inverse_transform(y_pred)

        return y_pred
